Package: pairscale
Type: Package
Title: Pairwise Rescaling of Numeric Matrices
Version: 1.0
Authors@R: 
    person(given = "Frank",
           family = "Koopmans",
           role = c("aut", "cre"),
           email = "ftwkoopmans@gmail.com",
           comment = c(ORCID = "0000-0002-4973-5732"))
Description: 
    Normalization of numerical matrices by minimizing the mean/median/mode difference between all column pairs.
URL: https://github.com/ftwkoopmans/pairscale/
License: AGPL (>= 3)
Encoding: UTF-8
SystemRequirements: C++20
Depends: R (>= 4.1.0)
Imports: Rcpp (>= 1.0.5)
Suggests: testthat (>= 3.0.0)
LinkingTo: Rcpp, RcppArmadillo (>= 14.7)
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Language: en-US
NeedsCompilation: yes
Packaged: 2026-05-13 14:12:22 UTC; frank
Author: Frank Koopmans [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4973-5732>)
Maintainer: Frank Koopmans <ftwkoopmans@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-19 07:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-19 10:27:31 UTC; unix
Archs: pairscale.so.dSYM

# Generated by roxygen2: do not edit by hand

export(pairdiff_madmean)
export(pairdiff_mean)
export(pairdiff_median)
export(pairdiff_mode)
export(pairdiff_trimmedmean)
export(pairscale_madmean)
export(pairscale_mean)
export(pairscale_median)
export(pairscale_mode)
export(pairscale_trimmedmean)
export(solve_graph_laplacian)
export(vector_madmean)
export(vector_mean)
export(vector_median)
export(vector_mode)
export(vector_trimmedmean)
importFrom(Rcpp,sourceCpp)
useDynLib(pairscale, .registration=TRUE)

pairscale is a highly optimized and robust implementation for pairwise
rescaling that supports various metrics of central tendency, including
the mode.

Pairwise normalization consists of two steps. First, for every unique
pair of columns in the input data matrix, a single measure of central
tendency (e.g. mean, median, mode) is computed over the difference in
their values (i.e. all values in column *i* minus all values in column
*j*). The result is a matrix *M* where columns and rows represent
columns in the input data matrix and values represent their difference
(or “distance”). Second, we find the optimal scaling factors for each
column in the input data matrix such that values in this matrix M are
minimized.

 

**License:**

[![License: AGPL
v3](https://img.shields.io/badge/License-AGPL_v3-blue.svg)](https://www.gnu.org/licenses/agpl-3.0)

**Citation:**

*manuscript under review*

# Installation

## 1) Installing R + tools for compiling

### Windows

- R version 4.1 (or higher) is required on Windows
  - if you already have R installed, check the current version using the
    R command: `R.version.string`
  - if you need to install or upgrade R
    - download the latest/current version from
      <https://cran.r-project.org/bin/windows/base/>
- RTools is required
  - check RTools is installed using the R command;
    `pkgbuild::check_rtools()` (assuming you have the “pkgbuild” package
    installed)
  - if not, get the RTools installer matching your R version from
    <https://cran.r-project.org/bin/windows/Rtools/>
- Reboot your computer (not mandatory, but this’ll avoid obscure
  bugs/issues)
- Download and install the RStudio desktop IDE. Current link;
  <https://posit.co/download/rstudio-desktop>

### MacOS

- R version 4.3 (or higher) is required on MacOS
  - if you already have R installed, check the current version using the
    R command: `R.version.string`
  - if you need to install or upgrade R
    - download the latest/current version from
      <https://mac.r-project.org> . For “apple silicon” MacBooks
      (approx. 2021 or newer), you’ll need
      “R-`<current version>`-arm64.pkg”
    - don’t use the devel/unstable
- Download and install the RStudio desktop IDE. Current link;
  <https://posit.co/download/rstudio-desktop>
- Install the toolchain for compiling code
  - start RStudio (if you previously opened it, close and start it to
    get a ‘fresh session’)
  - issue the following R statements (documentation @
    <https://github.com/coatless-mac/macrtools> )

``` r
# install the macrtools R package
install.packages("remotes")
remotes::install_github("coatless-mac/macrtools")

# non-interactive installation of Xcode CLI
macrtools::xcode_cli_install() 

# install gfortran
macrtools::gfortran_install()

# other binaries required for compiling R
macrtools::recipes_binary_install('r-base-dev')

# install OpenMP and configure Makevars
macrtools::openmp_install()
```

Potential issues;

- `Directory of lock file does not exist: /Users/<username>/Library/Caches/org.R-project.R/R/pkgcache/pkg`
  during installation suggests you lack write permission to (some part
  of) this path. Solution; check that this path exists (create if
  missing) and give your current OS user write permission (at least to
  the last 4 directories in this path)
- `ld: library not found for -lgfortran` if this is among your error
  messages, the Fortran/gfortran compiler is not installed
- `Failed to build source package <name>` –\> either Xcode/Fortran is
  not installed, or their version doesn’t line up with your current R
  version

### Linux

- R version 4.1.0 (or higher) and toolchains for compilation are
  required

## 2) Install the pairscale R package

pairscale is available on CRAN and can be installed using;
`install.packages("pairscale")`

- alternatively, install from this GitHub repository instead of CRAN;
  `remotes::install_github("ftwkoopmans/pairscale")`

On Windows, if you are prompted to install/upgrade RcppArmadillo from
source or use a precompiled binary, select “from source”.

# Quickstart

- goal: rescale all columns based on pairwise differences such that
  differences (between columns) are minimized
- input: typically a log2 transformed protein or gene abundance matrix
- this example: a tiny data matrix where we’ll use the median-difference
  between columns as a metric for rescaling/normalizing

``` r
library(pairscale)
# minimal example; a tiny data matrix
# note how the values in each columns are shifted by a constant
# rescaling should make all columns nearly identical
mat = cbind(
  c(1,2,3,4,5),
  c(2,3,4,5,6),
  c(1.1,2.1,3.1,4.1,5.1),
  c(2,3,4,5,9)
)
print(mat) # input data / prior to normalization
```

    ##      [,1] [,2] [,3] [,4]
    ## [1,]    1    2  1.1    2
    ## [2,]    2    3  2.1    3
    ## [3,]    3    4  3.1    4
    ## [4,]    4    5  4.1    5
    ## [5,]    5    6  5.1    9

``` r
# mat is rescaled and modified in-place by pairscale
# returned variable s represents the applied scaling factors
s = pairscale::pairscale_median(mat, min_value_count = 3)
print(mat) # normalized data
```

    ##       [,1]  [,2]  [,3]  [,4]
    ## [1,] 1.525 1.525 1.525 1.525
    ## [2,] 2.525 2.525 2.525 2.525
    ## [3,] 3.525 3.525 3.525 3.525
    ## [4,] 4.525 4.525 4.525 4.525
    ## [5,] 5.525 5.525 5.525 8.525

Above code rescales matrix `mat` considering all column pairs. But if
your dataset has clusters/groups, you can also normalize within
cluster/group and then normalize between groups using the `clusters`
parameter.

``` r
mat_groups = c(1,1,2,2) # provided groups/clusters must be positive integer values
s = pairscale::pairscale_median(mat, min_value_count = 3, clusters = mat_groups)
?pairscale::pairscale_median # see further @ function documentation
```

## a simulated dataset

For many real-world proteomics/transcriptomics datasets,
pairwise-differences between samples have an approximately normally
distributed null distribution (or more generally, a distribution where
the mode is at the center of the real null). If there are sufficient
data points, e.g. at least 50 rows in the input matrix (preferably
hundreds), using the mode is a robust metric for determining the
relative rescaling of each pair of samples/columns.

- this example: simulate a data matrix with 10 columns/samples, 1000
  rows/proteins and a substantial set of missing values for low-abundant
  rows/proteins in the first 5 samples
- expectation: there are 2 cohorts of distributions, i.e. full
  distribution and a subset lacking low values, however all overlapping
  values between columns should be similar post normalization
- observations:
  - mode normalization is robust and yields expected outcome
  - many data pipelines and algorithms (e.g. directLFQ) normalize
    between samples using the median value per sample as a reference
    value (‘median shift’ approach). This approach is inaccurate for
    datasets such as generated in this example!

``` r
library(pairscale)
set.seed(123) # fix random seed for reproducibility
baseline = rnorm(1000, mean = 0, sd = 1) # baseline values for each row/protein
mat = matrix(NA, nrow = 1000, ncol = 10) # simulated data matrix
simulated_scaling_factors = runif(n = 10, min = -3, max = 3) # randomly rescale each column
for(j in 1:10) {
  # values in column j; baseline values + a bit of random noise per row + mock scaling factor
  mat[,j] = baseline + runif(n = 1000, min = -0.1, max = 0.1) + simulated_scaling_factors[j]
  # in the first 5 samples, remove the lowest 25% of values (e.g. strong censoring)
  # as a result, the distributions between the first and second set of samples are distinct
  if(j <= 5) {
    mat[head(order(mat[,j], decreasing = FALSE), n = 250),j] = NA
  }
}

# QC: plot the data distributions
plot(NA, xlim = c(-4, 4), ylim = c(0, 0.75), type = "n", xlab = "value", ylab = "density", main = "INPUT DATA\nblack = full distribution in columns 6:10\nred = subset in columns 1:5")
for(j in 1:ncol(mat)) {
  lines(density(mat[,j], na.rm = TRUE), col = 1 + (j <= 5))
}
```

![](README_files/figure-gfm/unnamed-chunk-2-1.png)<!-- -->

``` r
# use pairscale to apply mode normalization
# variable 'mat' is updated by reference (no overhead of creating a copy of the data matrix)
# returned variable represents the applied scaling factors
estimated_scaling_factors = pairscale::pairscale_mode(mat, min_value_count = 25)

# for reference, compute scaling factors using the commonly used 'median shift' approach
mat_column_medians = apply(mat, 2, median, na.rm = TRUE)

# compare simulated differences in 'sample loading' against estimated normalization factors
# 1) pairscale recapitulates expected scaling factors; minor imprecision due to simulated random noise
# 2) column medians are not proper scaling factors for this simulated dataset;
#     'median shift' works for homogenous datasets but fails when sample distributions are dissimilar
print(cbind(
 "simulated" = simulated_scaling_factors - mean(simulated_scaling_factors), 
 "pairscale estimated" = estimated_scaling_factors - mean(estimated_scaling_factors),
 "column medians" = mat_column_medians - mean(mat_column_medians) 
))
```

    ##        simulated pairscale estimated column medians
    ##  [1,] -0.7263389          -0.7242676      0.1428896
    ##  [2,] -0.8172877          -0.8200412      0.1537885
    ##  [3,] -0.7893005          -0.7860611      0.1443449
    ##  [4,]  1.4022228           1.4086442      0.1493479
    ##  [5,]  1.2725810           1.2661394      0.1703957
    ##  [6,]  2.0136738           2.0077481     -0.1477401
    ##  [7,]  1.0001545           1.0012721     -0.1647724
    ##  [8,] -1.3503225          -1.3459117     -0.1452875
    ##  [9,] -1.6520049          -1.6510816     -0.1363727
    ## [10,] -0.3533776          -0.3564404     -0.1665937

``` r
plot(NA, xlim = c(-4, 4), ylim = c(0, 0.75), type = "n", xlab = "value", ylab = "density", main = "NORMALIZED\nblack = full distribution in columns 6:10\nred = subset in columns 1:5")
for(j in 1:ncol(mat)) {
  lines(density(mat[,j], na.rm = TRUE), col = 1 + (j <= 5))
}
```

![](README_files/figure-gfm/unnamed-chunk-2-2.png)<!-- -->

``` r
# visualize the normalized data matrix to illustrate all rows are similar except those set to NA
heatmap(mat[order(baseline),], Rowv = NA, Colv = NA, scale = "none")
```

![](README_files/figure-gfm/unnamed-chunk-2-3.png)<!-- -->

# benchmark computation speed

Here we evaluate the computation speed of pairscale functions that
compute measures of central tendency for a given numeric vector;

``` r
library(pairscale)
if(!requireNamespace("rbenchmark", quietly = TRUE)) install.packages("rbenchmark")

# find the mode using base R density function
r_density_mode = function(x, trim = 0) {
  if(trim > 0) {
    q = quantile(x, probs = c(trim, 1-trim), na.rm = TRUE)
    x = x[is.finite(x) & x > q[1] & x < q[2]]
  }
  d = stats::density(x, adjust = 1, n = 512, bw = "nrd", na.rm = TRUE)
  return(d$x[which.max(d$y)])
}

# 100k vector + 1 extreme value
x = c(rnorm(100000, mean = 0, sd = 1), 13)

# use rbenchmark package to time pairscale functions for computing measures of central tendency
# benchmark results shown here were generated on a MacBook Pro with a Apple M4 Max chip under R 4.4.2
rbenchmark::benchmark(
  "R trimmedmean" = {d = mean(x, trim = 0.2, na.rm=T)},
  "R mode" = {d = r_density_mode(x)},
  "R mode trimmed" = {d = r_density_mode(x, trim = 0.025)},
  "R median" = {d = median(x, na.rm=T)},
  "vector_median" = {d = vector_median(x)},
  "vector_trimmedmean" = {d = vector_trimmedmean(x, trim = 0.2)},
  "vector_mode(fastest 200)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 200, bandwidth_method = "nrd_fastest")},
  "vector_mode(fast 200)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 200, bandwidth_method = "nrd_fast")},
  "vector_mode(200)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 200, bandwidth_method = "nrd")},
  "vector_mode(fastest 512)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 512, bandwidth_method = "nrd_fastest")},
  "vector_mode(fast 512)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 512, bandwidth_method = "nrd_fast")},
  "vector_mode(512)" = {d = vector_mode(x, min_value_count = 1, adjust = 1, kernel_width_in_sd = 3, n_bins = 512, bandwidth_method = "nrd")},
  replications = 1000, order = "relative", columns = c("test", "replications", "elapsed", "relative", "user.self", "sys.self")
)
```

Results on a MacBook Pro:

    #                        test replications elapsed relative user.self sys.self
    # 7  vector_mode(fastest 200)         1000   0.195    1.000     0.194    0.000
    # 10 vector_mode(fastest 512)         1000   0.203    1.041     0.202    0.000
    # 5             vector_median         1000   0.474    2.431     0.471    0.003
    # 6        vector_trimmedmean         1000   0.782    4.010     0.782    0.000
    # 4                  R median         1000   0.984    5.046     0.828    0.155
    # 11    vector_mode(fast 512)         1000   0.999    5.123     0.998    0.001
    # 8     vector_mode(fast 200)         1000   1.059    5.431     1.058    0.001
    # 12         vector_mode(512)         1000   1.347    6.908     1.345    0.002
    # 9          vector_mode(200)         1000   1.453    7.451     1.451    0.002
    # 1             R trimmedmean         1000   1.479    7.585     1.265    0.213
    # 2                    R mode         1000   2.012   10.318     1.831    0.180
    # 3            R mode trimmed         1000   3.749   19.226     3.294    0.454

Results on a Windows workstation:

    #                        test replications elapsed relative user.self sys.self
    # 10 vector_mode(fastest 512)         1000    0.38    1.000      0.38     0.00
    # 7  vector_mode(fastest 200)         1000    0.43    1.132      0.43     0.00
    # 5             vector_median         1000    0.72    1.895      0.72     0.00
    # 6        vector_trimmedmean         1000    1.14    3.000      1.14     0.00
    # 11    vector_mode(fast 512)         1000    1.30    3.421      1.30     0.00
    # 8     vector_mode(fast 200)         1000    1.33    3.500      1.33     0.00
    # 12         vector_mode(512)         1000    1.75    4.605      1.75     0.00
    # 9          vector_mode(200)         1000    1.76    4.632      1.76     0.00
    # 4                  R median         1000    2.09    5.500      2.01     0.08
    # 1             R trimmedmean         1000    4.24   11.158      3.14     1.11
    # 2                    R mode         1000    4.86   12.789      4.06     0.80
    # 3            R mode trimmed         1000    7.71   20.289      7.39     0.33

## a note for R package developers

If you download and open the pairscale.Rproj file and use
`devtools::load_all()`, instead of going the usual route of
`install.packages("pairscale"); library(pairscale)`, the above
benchmarks will underperform because devtools will override compiler
flags (setting e.g. -O0). To disable this, and thus only use your
~/.R/Makevars (if available), run this code instead;

``` r
Sys.setenv("PKG_BUILD_EXTRA_FLAGS" = "false")
devtools::load_all(recompile = TRUE)
```

# Copyright (C) 2026 Frank Koopmans
#
# This file is part of pairscale.
#
# pairscale is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pairscale is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


#' pairscale package declaration
#' @name pairscale
#' @useDynLib pairscale, .registration=TRUE
#' @keywords internal
#' @importFrom Rcpp sourceCpp
"_PACKAGE"

#' cleanup Rcpp code
#'
#' @param libpath library path
#' @noRd
.onUnload = function(libpath = NULL) {
  library.dynam.unload("pairscale", libpath)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @title Compute mean value of a vector, with optional filtering for N datapoints
#' @description Compute measure of central tendency using efficient C++ code
#' @name vector_mean
#' @inheritParams vector_mode
#' @inheritParams pairscale_mode
#' @return a single numeric value representing the mean
#' @export
vector_mean <- function(x, min_value_count = 1L, na_mode = "check") {
    .Call(`_pairscale_vector_mean`, x, min_value_count, na_mode)
}

#' @title Compute median value of a vector, with optional filtering for N datapoints
#' @description Compute measure of central tendency using efficient C++ code
#' @name vector_median
#' @inheritParams vector_mode
#' @inheritParams pairscale_mode
#' @return a single numeric value representing the median
#' @export
vector_median <- function(x, min_value_count = 1L, na_mode = "check") {
    .Call(`_pairscale_vector_median`, x, min_value_count, na_mode)
}

#' @title Compute trimmed mean value of a vector, with optional filtering for N datapoints
#' @description Compute measure of central tendency using efficient C++ code
#' @name vector_trimmedmean
#' @inheritParams vector_mode
#' @inheritParams pairscale_mode
#' @inheritParams pairscale_trimmedmean
#' @return a single numeric value representing the trimmed mean
#' @export
vector_trimmedmean <- function(x, min_value_count = 1L, trim = 0.2, na_mode = "check") {
    .Call(`_pairscale_vector_trimmedmean`, x, min_value_count, trim, na_mode)
}

#' @title Compute MAD-trimmed mean value of a vector, with optional filtering for N datapoints
#' @description Compute measure of central tendency using efficient C++ code
#' @name vector_madmean
#' @inheritParams vector_mode
#' @inheritParams pairscale_mode
#' @inheritParams pairscale_madmean
#' @return a single numeric value representing the MAD-trimmed mean
#' @export
vector_madmean <- function(x, min_value_count = 1L, threshold_std = 3, na_mode = "check") {
    .Call(`_pairscale_vector_madmean`, x, min_value_count, threshold_std, na_mode)
}

#' @title Compute mode of a vector, with optional filtering for N datapoints
#' @description Compute measure of central tendency using efficient C++ code
#' @name vector_mode
#' @inheritParams pairscale_mode
#' @param x numeric input vector, may contain non-finite values (removed if `na_mode` is left to default)
#' @return a single numeric value representing the mode
#' @export
vector_mode <- function(x, min_value_count = 3L, n_bins = 512L, adjust = 1, kernel_width_in_sd = 3, bandwidth_method = "nrd", mode_frac_maxdens = 1, na_mode = "check") {
    .Call(`_pairscale_vector_mode`, x, min_value_count, n_bins, adjust, kernel_width_in_sd, bandwidth_method, mode_frac_maxdens, na_mode)
}

#' @title Distance matrix between columns of a matrix using their mean differences
#' @description Pairwise differences between all columns in a matrix
#' @name pairdiff_mean
#' @inheritParams pairscale_mode
#' @inheritParams pairdiff_mode
#' @return a N x N numeric matrix (where N is number of column in input `x`) representing the mean difference between each column
#' @export
pairdiff_mean <- function(x, cols = NULL, min_value_count = 3L, na_mode = "check") {
    .Call(`_pairscale_pairdiff_mean`, x, cols, min_value_count, na_mode)
}

#' @title Distance matrix between columns of a matrix using their trimmed median differences
#' @description Pairwise differences between all columns in a matrix
#' @name pairdiff_median
#' @inheritParams pairscale_mode
#' @inheritParams pairdiff_mode
#' @return a N x N numeric matrix (where N is number of column in input `x`) representing the median difference between each column
#' @export
pairdiff_median <- function(x, cols = NULL, min_value_count = 3L, na_mode = "check") {
    .Call(`_pairscale_pairdiff_median`, x, cols, min_value_count, na_mode)
}

#' @title Distance matrix between columns of a matrix using their trimmed mean differences
#' @description Pairwise differences between all columns in a matrix
#' @name pairdiff_trimmedmean
#' @inheritParams pairscale_mode
#' @inheritParams pairdiff_mode
#' @inheritParams pairscale_trimmedmean
#' @return a N x N numeric matrix (where N is number of column in input `x`) representing the trimmed mean difference between each column
#' @export
pairdiff_trimmedmean <- function(x, cols = NULL, min_value_count = 3L, trim = 0.2, na_mode = "check") {
    .Call(`_pairscale_pairdiff_trimmedmean`, x, cols, min_value_count, trim, na_mode)
}

#' @title Distance matrix between columns of a matrix using their MAD-trimmed mean differences
#' @description Pairwise differences between all columns in a matrix
#' @name pairdiff_madmean
#' @inheritParams pairscale_mode
#' @inheritParams pairdiff_mode
#' @inheritParams pairscale_madmean
#' @return a N x N numeric matrix (where N is number of column in input `x`) representing the MAD-trimmed mean difference between each column
#' @export
pairdiff_madmean <- function(x, cols = NULL, min_value_count = 3L, threshold_std = 3, na_mode = "check") {
    .Call(`_pairscale_pairdiff_madmean`, x, cols, min_value_count, threshold_std, na_mode)
}

#' @title Distance matrix between columns of a matrix using their mode differences
#' @description Pairwise differences between all columns in a matrix
#' @name pairdiff_mode
#' @inheritParams pairscale_mode
#' @param cols optionally, provide an integer vector with column indices (in `x`) that should be used (these should be 1-based indices as per usual in R). Or set to `NULL` or `integer()` to use all columns
#' @return a N x N numeric matrix (where N is number of column in input `x`) representing the mode difference between each column
#' @export
pairdiff_mode <- function(x, cols = NULL, min_value_count = 3L, n_bins = 512L, adjust = 1, kernel_width_in_sd = 3, bandwidth_method = "nrd", mode_frac_maxdens = 1, na_mode = "check") {
    .Call(`_pairscale_pairdiff_mode`, x, cols, min_value_count, n_bins, adjust, kernel_width_in_sd, bandwidth_method, mode_frac_maxdens, na_mode)
}

#' @title Normalize matrix columns using their mode differences
#' @description Pairwise normalization of columns in a matrix, using the mode to define pairwise distances between columns
#' @name pairscale_mode
#' @param x a numeric matrix that needs to be normalized. This variable is changed by reference! i.e. after this function, the original variable is updated
#' @param clusters an integer vector, of same length as `ncol(x)`, that describes cluster identifiers. These values must describe a continuous set of integers between 1 and the total number of unique clusters. For example, a 6 column matrix with 3 columns from some experimental condition followed by 3 columns from another condition would be `c(1,1,1, 2,2,2)`. To disable mode-between rescaling, i.e. treat the entire matrix as one cluster, provide a vector with ones.
#' @param min_value_count the minimum number of values overlapping between a pair of columns (if there are fewer overlapping values, respective rescaling is not computed)
#' @param n_bins grid size for density computation (the resolution / number of data points to use for binning column diffs)
#' @param adjust bandwidth adjust factor for density computation
#' @param kernel_width_in_sd maximum distance in standard deviations at which we'll include data points for the Gaussian kernal. Typically 3 or 4
#' @param na_mode string value that indicated how should we should deal with NA values (default). "check" = test if NA values are present and if so, remove these. "present" = you already know NA values are present so we can skip the check for NA values for efficiency. "unchecked" = you guarantee that there are no NA values in input, thus we'll use the fastest code paths that skip over any NA checks downstream; ONLY select this if you are sure there are no NA/NaN/Inf/-Inf values !
#' @param bandwidth_method method in which this function computes bandwidth and optionally trims the data prior to binning. "nrd" is the robust, safe default. "nrd_fast" is faster and yields similar results for most distributions. Use "nrd_fastest" only when all pairwise distances are known to be near gaussian (i.e. no strong outliers and sd() is a reliable metric). "nrd_subset" is an experimental option that may be removed, it is fast but heavily favors symmetric distributions and is thus biased !
#'
#'   Valid options:
#'   * `"nrd_fastest"` ; use all data points, for bandwidth don't use IQR. Sensitive to outliers !
#'   * `"nrd_fast"` ; trim to quantiles 0.05 and 0.95, then use sd over these data points to compute bandwidth. Not as robust as the original nrd method, which also considers IQR, but much faster.
#'   * `"nrd"` ; adapted version of nrd; instead of IQR and sd we use IQR and sd over data trimmed at quantiles 0.05 and 0.95. Since we already sorted the data, computing sd on trimmed subset is 'free'
#' @param mode_frac_maxdens set to 1 to return the x-coordinate where the density is highest (mode). Setting this to a value < 1 will make this function compute not the mode, but the mean (x) value of the density where the density is some fraction higher than the maximum density. Typical value; 1. Optionally, set to 0.9 or 0.8 for possibly more robust center-finding, depending in your data distribution. Must not be smaller than 0.1 but recommended to never be lower than 0.7
#' @param niter_irls number of iterations to increase robustness (set to zero to disable)
#' @return a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
#' @export
pairscale_mode <- function(x, clusters = NULL, min_value_count = 3L, n_bins = 512L, adjust = 1, kernel_width_in_sd = 3, bandwidth_method = "nrd", mode_frac_maxdens = 1, niter_irls = 50L, na_mode = "check") {
    .Call(`_pairscale_pairscale_mode`, x, clusters, min_value_count, n_bins, adjust, kernel_width_in_sd, bandwidth_method, mode_frac_maxdens, niter_irls, na_mode)
}

#' @title Normalize matrix columns using their median differences
#' @description Pairwise normalization of columns in a matrix, using the median to define pairwise distances between columns
#' @name pairscale_median
#' @inheritParams pairscale_mode
#' @return a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
#' @export
pairscale_median <- function(x, clusters = NULL, min_value_count = 3L, niter_irls = 50L, na_mode = "check") {
    .Call(`_pairscale_pairscale_median`, x, clusters, min_value_count, niter_irls, na_mode)
}

#' @title Normalize matrix columns using their mean differences
#' @description Pairwise normalization of columns in a matrix, using the mean to define pairwise distances between columns
#' @name pairscale_mean
#' @inheritParams pairscale_mode
#' @return a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
#' @export
pairscale_mean <- function(x, clusters = NULL, min_value_count = 3L, niter_irls = 50L, na_mode = "check") {
    .Call(`_pairscale_pairscale_mean`, x, clusters, min_value_count, niter_irls, na_mode)
}

#' @title Normalize matrix columns using their trimmed-mean differences
#' @description Pairwise normalization of columns in a matrix, using the trimmed-mean to define pairwise distances between columns
#' @name pairscale_trimmedmean
#' @inheritParams pairscale_mode
#' @param trim amount of trim to apply to both the lower- and upper-parts of a vector before computing the mean. 0 indicates no trim, 0.5 indicates 100% trim (i.e. 50% of data on both sides) so that value is out of bounds. Typically set to 0.1-0.3
#' @return a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
#' @export
pairscale_trimmedmean <- function(x, clusters = NULL, min_value_count = 3L, trim = 0.2, niter_irls = 50L, na_mode = "check") {
    .Call(`_pairscale_pairscale_trimmedmean`, x, clusters, min_value_count, trim, niter_irls, na_mode)
}

#' @title Normalize matrix columns using their MAD-trimmed mean differences
#' @description Pairwise normalization of columns in a matrix, using the MAD-trimmed mean to define pairwise distances between columns
#' @name pairscale_madmean
#' @inheritParams pairscale_mode
#' @param threshold_std ratio of MAD a value has to be away from the median to be considered an outlier (and thus removed/ignored). Note that the MAD thresholds are inclusive, i.e. values at +/- threshold_std*MAD from median are included
#' @return a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
#' @export
pairscale_madmean <- function(x, clusters = NULL, min_value_count = 3L, threshold_std = 3, niter_irls = 50L, na_mode = "check") {
    .Call(`_pairscale_pairscale_madmean`, x, clusters, min_value_count, threshold_std, niter_irls, na_mode)
}

#' @title graph Laplacian approach to finding normalization factors
#' @description find normalization factors for a given distance matrix computed with e.g. `pairdiff_median()`. For increased robustness, this function offers iterative reweighted improvement of the initial estimate.
#' @name solve_graph_laplacian
#' @param M skew-symmetric input matrix, generated with e.g. `pairdiff_median()`
#' @param niter_irls refine the initial estimate using N additional iterative reweighted least squares loops for robust graph laplacian
#' @examples
#' # toy example
#' x = cbind(
#'   c(1,2,3,4),
#'   c(2,3,4,9),
#'   c(1,2,4,5),
#'   c(1,0,1,0)
#' )
#' # compute pairwide median difference between all columns
#' M = pairscale::pairdiff_median(x)
#' # solve matrix M to find scaling factors, without and with reweighting
#' s1 = pairscale::solve_graph_laplacian(M, niter_irls = 0)
#' s2 = pairscale::solve_graph_laplacian(M, niter_irls = 10)
#' # rescaled matrices; only the robust variant correctly aligns columns 1 and 2
#' t(t(x) - s1)
#' t(t(x) - s2)
#' @return a numeric vector of length `ncol(M)` that contains scaling factors for `M`
#' @export
solve_graph_laplacian <- function(M, niter_irls = 1L) {
    .Call(`_pairscale_solve_graph_laplacian`, M, niter_irls)
}

# data: apply constant scaling factor to each column
# post normalization, all values on each rows should be the same, i.e. the diff computed here should be zero
for(iter in 1:10) {

  tmp = rnorm(100)
  x = cbind(tmp, tmp + 3, tmp - 1)
  l = list()

  y = x + 1  # make a deep copy / copy-on-modify
  s = pairscale_mean(y)
  l$pairscale_mean = y

  y = x + 1
  s = pairscale_median(y)
  l$pairscale_median = y

  y = x + 1
  s = pairscale_madmean(y)
  l$pairscale_madmean = y

  y = x + 1
  s = pairscale_trimmedmean(y)
  l$pairscale_trimmedmean = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd_fastest")
  l$pairscale_mode_fastest = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd_fast")
  l$pairscale_mode_fast = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd")
  l$pairscale_mode = y
  # boxplot(x + 1, main = "before")
  # boxplot(y, main = "after - mode should be equal, median may not be !")


  for(function_name in names(l)) {
    y = l[[function_name]]

    test_that(paste0("3 column fixed scaling, ", function_name, ", columns 1,2"), {
      expect_equal(y[,1], y[,2], tolerance = 0.001)
    })
    test_that(paste0("3 column fixed scaling, ", function_name, ", columns 1,3"), {
      expect_equal(y[,1], y[,3], tolerance = 0.001)
    })
  }

  rm(tmp, x, l, s)
}


r_density_mode = function(x) {
  y = na.omit(x)
  d = stats::density(y[y > quantile(y, probs=0.05) & y < quantile(y, probs=0.95)], bw = "nrd", adjust = 0.9)
  return(d$x[which.max(d$y)])
}

for(iter in 1:10) {

  l = list()
  tmp = rnorm(1000)
  tmp1 = c(tmp + runif(length(tmp), min = -0.1, max = 0.2), rnorm(120, mean = 2, sd = 1))
  tmp2 = c(tmp + runif(length(tmp), min = -0.1, max = 0.2), rnorm(120, mean = 2, sd = 1))
  x = cbind(tmp1, tmp1 + 3, tmp1 - 1, tmp2 + 3, tmp2 - 2, tmp2 + 1)
  x_clusters = c(1,1,1, 9,9,9)
  # plot(density(tmp1)); lines(density(tmp2), col=2)
  # plot(density(x[,1] - x[,5]))


  y = x + 1  # make a deep copy / copy-on-modify
  s = pairscale_mean(y)
  l$pairscale_mean = y

  y = x + 1
  s = pairscale_median(y)
  l$pairscale_median = y

  y = x + 1
  s = pairscale_madmean(y)
  l$pairscale_madmean = y

  y = x + 1
  s = pairscale_trimmedmean(y)
  l$pairscale_trimmedmean = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd_fastest")
  l$pairscale_mode_fastest = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd_fast")
  l$pairscale_mode_fast = y

  y = x + 1
  s = pairscale_mode(y, bandwidth_method = "nrd")
  l$pairscale_mode = y
  # boxplot(x + 1, main = "before")
  # boxplot(y, main = "after - mode should be equal, median may not be !")


  # mean scaling yields errors on some random datasets, it's very sensitive to outliers
  # we do run it while unit testing to ensure it doesn't generate errors
  y = x + 1
  s = pairscale_mean(y, clusters = x_clusters)
  # l$pairscale_mean_withclusters = y

  y = x + 1
  s = pairscale_median(y, clusters = x_clusters)
  l$pairscale_median_withclusters = y

  y = x + 1
  s = pairscale_madmean(y, clusters = x_clusters)
  l$pairscale_madmean_withclusters = y

  y = x + 1
  s = pairscale_trimmedmean(y, clusters = x_clusters)
  l$pairscale_trimmedmean_withclusters = y

  y = x + 1
  s = pairscale_mode(y, clusters = x_clusters, bandwidth_method = "nrd_fastest")
  l$pairscale_mode_fastest_withclusters = y

  y = x + 1
  s = pairscale_mode(y, clusters = x_clusters, bandwidth_method = "nrd_fast")
  l$pairscale_mode_fast_withclusters = y

  y = x + 1
  s = pairscale_mode(y, clusters = x_clusters, bandwidth_method = "nrd")
  l$pairscale_mode_withclusters = y


  for(function_name in names(l)) {
    y = l[[function_name]]
    test_that(paste0("minor noise, ", function_name, ", columns 1,2"), {
      expect_equal(y[,1], y[,2], tolerance = 0.001)
    })
    test_that(paste0("minor noise, ", function_name, ", columns 1,3"), {
      expect_equal(y[,1], y[,3], tolerance = 0.001)
    })
    test_that(paste0("minor noise, ", function_name, ", columns 4,5"), {
      expect_equal(y[,4], y[,5], tolerance = 0.001)
    })
    test_that(paste0("minor noise, ", function_name, ", zero mode in 1,4 post normalization"), {
      expect_equal(r_density_mode(y[,1] - y[,4]), 0, tolerance = 0.1)
    })

  }



  rm(tmp, tmp1, tmp2, x, s)
}

# tricky matrix to normalize with "mode" approach, testing the density function
# in this example, the ratio between columns can be a constant value, or zero
x = cbind(1:5, 1:5 + 0.1, 1:5 + 0.1, 1:5 - 3, 1:5 - 3.3)
x_expect = rbind(
  c(0.0,-0.1,-0.1,3.0,3.3),
  c(0.1,0.0,0.0,3.1,3.4),
  c(0.1,0.0,0.0,3.1,3.4),
  c(-3.0,-3.1,-3.1,0.0,0.3),
  c(-3.3,-3.4,-3.4,-0.3,0.0)
)



for(colspec in list(1:2, 3:4, c(5,2))) {

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_median() vs expected values"), {
    expect_equal(pairdiff_median(x, cols = colspec, min_value_count = 3, na_mode = "present"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mean() vs expected values"), {
    expect_equal(pairdiff_mean(x, cols = colspec, min_value_count = 3, na_mode = "present"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_trimmedmean() vs expected values"), {
    expect_equal(pairdiff_trimmedmean(x, cols = colspec, min_value_count = 3, na_mode = "present"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_madmean() vs expected values"), {
    expect_equal(pairdiff_madmean(x, cols = colspec, min_value_count = 3, threshold_std = 3, na_mode = "present"), x_expect[colspec,colspec])
  })


  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(fastest, na=present) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fastest"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(fastest, na=unchecked) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "unchecked", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fastest"), x_expect[colspec,colspec])
  })


  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(fast, na=present) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fast"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(fast, na=unchecked) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "unchecked", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fast"), x_expect[colspec,colspec])
  })


  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(na=present) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd"), x_expect[colspec,colspec])
  })

  test_that(paste0("hardcoded data, subset of columns ", colspec[1],",",colspec[2]," ; pairdiff_mode(na=unchecked) vs expected values"), {
    expect_equal(pairdiff_mode(x, cols = colspec, min_value_count = 3, na_mode = "unchecked", n_bin = 100, adjust = 1, bandwidth_method = "nrd"), x_expect[colspec,colspec])
  })

}







### min_value_count=5 should remove pairwise scalings for sample 1 in this example
x = cbind(c(1:4, NA), 1:5 + 0.1, 1:5 + 0.1, 1:5 - 3, 1:5 - 3.3)
x_expect_na = matrix(FALSE, nrow = 5, ncol = 5)
x_expect_na[1,2:5] = TRUE
x_expect_na[2:5,1] = TRUE

test_that("hardcoded data, pairdiff_median() - min_value_count check", {
  expect_equal(is.na(pairdiff_median(x, min_value_count = 5, na_mode = "present")), x_expect_na)
})

test_that("hardcoded data, pairdiff_median() - min_value_count check", {
  expect_equal(is.na(pairdiff_mean(x, min_value_count = 5, na_mode = "present")), x_expect_na)
})

test_that("hardcoded data, pairdiff_trimmedmean() - min_value_count check", {
  expect_equal(is.na(pairdiff_trimmedmean(x, min_value_count = 5, na_mode = "present")), x_expect_na)
})

test_that("hardcoded data, pairdiff_madmean() - min_value_count check", {
  expect_equal(is.na(pairdiff_madmean(x, min_value_count = 5, threshold_std = 3, na_mode = "present")), x_expect_na)
})

test_that("hardcoded data, pairdiff_mode(fastest) - min_value_count check 1", {
  expect_equal(is.na(pairdiff_mode(x, min_value_count = 5, na_mode = "present", n_bins = 100, bandwidth_method = "nrd_fastest")), x_expect_na)
})

test_that("hardcoded data, pairdiff_mode(fast) - min_value_count check 1", {
  expect_equal(is.na(pairdiff_mode(x, min_value_count = 5, na_mode = "present", n_bins = 100, bandwidth_method = "nrd_fast")), x_expect_na)
})

test_that("hardcoded data, pairdiff_mode() - min_value_count check 1", {
  expect_equal(is.na(pairdiff_mode(x, min_value_count = 5, na_mode = "present", n_bins = 100, bandwidth_method = "nrd")), x_expect_na)
})








# with minval filtering, we can artifically create 2 blocks that have no overlap (considering min_value_count)
# -->> test that we get finite scaling factors and there is no runaway
x = cbind(c(1:4, NA), c(1:3 + 0.1, NA, NA), c(NA, NA, 1:3 + 0.1), c(NA, NA, 1:3 + 0.3))
x_expect = matrix(c(0,-0.1,NA,NA, 0.1,0,NA,NA, NA,NA,0,-0.2, NA,NA,0.2,0), ncol=4, byrow = T)

test_that("hardcoded data, partial column overlap, pairdiff_median()", {
  expect_equal(pairdiff_median(x, min_value_count = 3, na_mode = "present"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_median()", {
  expect_equal(pairdiff_mean(x, min_value_count = 3, na_mode = "present"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_trimmedmean()", {
  expect_equal(pairdiff_trimmedmean(x, min_value_count = 3, na_mode = "present"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_madmean()", {
  expect_equal(pairdiff_madmean(x, min_value_count = 3, threshold_std = 3, na_mode = "present"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_mode(fastest)", {
  expect_equal(pairdiff_mode(x, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fastest"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_mode(fast)", {
  expect_equal(pairdiff_mode(x, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd_fast"), x_expect)
})

test_that("hardcoded data, partial column overlap, pairdiff_mode()", {
  expect_equal(pairdiff_mode(x, min_value_count = 3, na_mode = "present", n_bin = 100, adjust = 1, bandwidth_method = "nrd"), x_expect)
})

r_vector_madmean = function(x, threshold_std = 3) {
  x = x[is.finite(x)]
  if(length(x) <= 3) return (median(x))
  mean(x[ x >= median(x) - threshold_std * mad(x) & x <= median(x) + threshold_std * mad(x) ])
}



test_that("vector_trimmedmean() deal with Inf, test 1", {
  expect_identical(is.na(vector_trimmedmean(c(-Inf), min_value_count = 1, trim = 0.2, na_mode = "check")), TRUE)
})
test_that("vector_trimmedmean() deal with NA, test 1", {
  expect_identical(is.na(vector_trimmedmean(c(NA), min_value_count = 3, trim = 0.2, na_mode = "check")), TRUE)
})
test_that("vector_trimmedmean() deal with NA, test 2", {
  expect_identical(is.na(vector_trimmedmean(c(NA, NA), min_value_count = 3, trim = 0.2, na_mode = "check")), TRUE)
})
test_that("vector_trimmedmean() deal with NA, test 3", {
  expect_identical(is.na(vector_trimmedmean(c(NA, NA, 1), min_value_count = 3, trim = 0.2, na_mode = "check")), TRUE)
})
test_that("vector_trimmedmean() deal with NA, test 4", {
  expect_identical(is.na(vector_trimmedmean(c(NA, NA, 1, 2), min_value_count = 3, trim = 0.2, na_mode = "check")), TRUE)
})
test_that("vector_trimmedmean() deal with NA, test 5", {
  expect_identical(vector_trimmedmean(c(NA, NA, 1, 2, 3), min_value_count = 3, trim = 0.2, na_mode = "check"), 2)
})
test_that("vector_trimmedmean() deal with NA, test 6", {
  expect_identical(vector_trimmedmean(c(NA, NA, 2), min_value_count = 1, trim = 0.2, na_mode = "check"), 2)
})



test_that("vector_madmean() deal with Inf, test 1", {
  expect_identical(is.na(vector_madmean(c(-Inf), min_value_count = 1, na_mode = "check")), TRUE)
})
test_that("vector_madmean() deal with NA, test 1", {
  expect_identical(is.na(vector_madmean(c(NA), min_value_count = 3, na_mode = "check")), TRUE)
})
test_that("vector_madmean() deal with NA, test 2", {
  expect_identical(is.na(vector_madmean(c(NA, NA), min_value_count = 3, na_mode = "check")), TRUE)
})
test_that("vector_madmean() deal with NA, test 3", {
  expect_identical(is.na(vector_madmean(c(NA, NA, 1), min_value_count = 3, na_mode = "check")), TRUE)
})
test_that("vector_madmean() deal with NA, test 4", {
  expect_identical(is.na(vector_madmean(c(NA, NA, 1, 2), min_value_count = 3, na_mode = "check")), TRUE)
})
test_that("vector_madmean() deal with NA, test 5", {
  expect_identical(vector_madmean(c(NA, NA, 1, 2, 3), min_value_count = 3, na_mode = "check"), 2)
})
test_that("vector_madmean() deal with NA, test 6", {
  expect_identical(vector_madmean(c(NA, NA, 2), min_value_count = 1, na_mode = "check"), 2)
})



test_that("vector_median() deal with Inf, test 1", {
  expect_identical(is.na(vector_median(c(-Inf), min_value_count = 1)), TRUE)
})
test_that("vector_median() deal with NA, test 1", {
  expect_identical(is.na(vector_median(c(NA), min_value_count = 3)), TRUE)
})
test_that("vector_median() deal with NA, test 2", {
  expect_identical(is.na(vector_median(c(NA, NA), min_value_count = 3)), TRUE)
})
test_that("vector_median() deal with NA, test 3", {
  expect_identical(is.na(vector_median(c(NA, NA, 1), min_value_count = 3)), TRUE)
})
test_that("vector_median() deal with NA, test 4", {
  expect_identical(is.na(vector_median(c(NA, NA, 1, 2), min_value_count = 3)), TRUE)
})
test_that("vector_median() deal with NA, test 5", {
  expect_identical(vector_median(c(NA, NA, 1, 2, 3), min_value_count = 3), 2)
})
test_that("vector_median() deal with NA, test 6", {
  expect_identical(vector_median(c(NA, NA, 2), min_value_count = 1), 2)
})




test_that("vector_trimmedmean() deal with vector of constants", {
  expect_identical(vector_trimmedmean(rep(1,5), min_value_count = 3, trim = 0.2, na_mode = "check"), 1)
})
test_that("vector_trimmedmean() deal with vector of constants", {
  expect_identical(vector_madmean(rep(1,5), min_value_count = 3, na_mode = "check"), 1)
})
test_that("vector_trimmedmean() deal with vector of constants", {
  expect_identical(vector_median(rep(1,5), min_value_count = 3), 1)
})



# hardcoded testcase; x = c(0.4200354, 0.4798072, -1.2270806)

for(iter in 1:100) {
  # array with random values, N = 1..100
  x = rnorm(sample.int(100, 1))
  # try a range of trim settings, from 0 (no trim) to 0.45
  for(iter_trim in (0:9 / 20)) {
    test_that(paste0("vector_trimmedmean() random data + NA iter=", iter, " trim=", iter_trim), {
      expect_equal(vector_trimmedmean(c(x, NA), trim = iter_trim, na_mode = "check"), mean(x, trim = iter_trim), tolerance = 10^-6)
    })

    test_that(paste0("vector_trimmedmean() random data iter=", iter, " trim=", iter_trim), {
      expect_equal(vector_trimmedmean(x, trim = iter_trim, na_mode = "check"), mean(x, trim = iter_trim), tolerance = 10^-6)
    })

  }


  test_that(paste0("vector_madmean() random data iter=", iter), {
    expect_equal(vector_madmean(x, min_value_count = 1, threshold_std = 3, na_mode = "check"), r_vector_madmean(x, threshold_std = 3), tolerance = 10^-6)
  })

  test_that(paste0("vector_median() random data iter=", iter), {
    expect_equal(vector_median(x, min_value_count = 1), median(x, na.rm = TRUE), tolerance = 10^-6)
  })

}

for(iter in 1:100) {
  x = runif(1, -100, 100)
  test_that("mode(fastest) deal with vector of constants - double", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd_fastest"), x)
  })
  test_that("mode(fast) deal with vector of constants - double", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd_fast"), x)
  })
  test_that("mode() deal with vector of constants - double", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd"), x)
  })

  x = sample.int(200, size = 1) - 100
  test_that("mode(fastest) deal with vector of constants - integer", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd_fastest"), x)
  })
  test_that("mode(fast) deal with vector of constants - integer", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd_fast"), x)
  })
  test_that("mode() deal with vector of constants - integer", {
    expect_identical(vector_mode(rep(x, 100), min_value_count = 3, n_bins = 200, adjust = 1, kernel_width_in_sd = 3, na_mode = "check", bandwidth_method = "nrd"), x)
  })
}


# instead of comparing the x coordinate,
# which can differ between distributions (shapes) and density settings such as bandwidth,
# we here compare distance in height of the density.
# e.g. if X is a noisy normal distribution with a "flat top" then minor changes in KDE
# will affect x-coordinates, while the objective calling of the mode (top of distribution) might be very similar.


for(iter in 1:100) {
  for(withna in c(FALSE, TRUE)) {
    # generate test data; normal distribution with 10k data points
    mu = runif(1, min = -10, max = 10)
    x = rnorm(10000, mean = mu, sd = 1)
    if(withna) {
      x = c(NA, x, Inf, -Inf)
    }
    # save(x, file = "~/temp/x_testval_density_fail.RData")
    # load("~/temp/x_testval_density_fail.RData")

    # density function in base R
    d = stats::density(x, adjust = 0.9, bw = 'nrd', na.rm = TRUE)
    mode_baser = d$x[which.max(d$y)]

    mode_pairscale_fastest = vector_mode(x, adjust = 0.9, bandwidth_method = "nrd_fastest")
    mode_pairscale_fast = vector_mode(x, adjust = 0.9, bandwidth_method = "nrd_fast")
    mode_pairscale = vector_mode(x, adjust = 0.9, bandwidth_method = "nrd")

    density_xy_baser = list(x = mode_baser, y = max(d$y))
    density_xy_pairscale_fastest = approx(d$x, d$y, xout = mode_pairscale_fastest) # linear interpolation by x-coord
    density_xy_pairscale_fast = approx(d$x, d$y, xout = mode_pairscale_fast) # linear interpolation by x-coord
    density_xy_pairscale = approx(d$x, d$y, xout = mode_pairscale)

    # validate that there is less than 1% difference in density between where
    # base R functions calls the mode and where our Rcpp implementations do.
    y_diff_frac_fastest = abs(density_xy_baser$y - density_xy_pairscale_fastest$y) / density_xy_baser$y
    y_diff_frac_fast    = abs(density_xy_baser$y - density_xy_pairscale_fast$y) / density_xy_baser$y
    y_diff_frac         = abs(density_xy_baser$y - density_xy_pairscale$y) / density_xy_baser$y

    test_that(paste0("mode(fastest) random data iter=", iter), {
      expect(y_diff_frac_fastest < 0.01, failure_message = "y_diff_frac_fast < 0.01")
    })
    test_that(paste0("mode(fast) random data iter=", iter), {
      expect(y_diff_frac_fast < 0.01, failure_message = "y_diff_frac_fast < 0.01")
    })
    test_that(paste0("mode() random data iter=", iter), {
      expect(y_diff_frac < 0.01, failure_message = "y_diff_fract_robust < 0.01")
    })

  }

}

# # in case a test fails, use this code to plot the distribution and respective x-values
# plot(stats::density(x, adjust = 0.9, na.rm = TRUE))
# abline(v = c(mode_baser, mode_pairscale_fastest, mode_pairscale_fast, mode_pairscale), col = 1:4, lwd = 2, lty = c(1,2,2,2))
# c(mu, mode_baser, mode_pairscale_fastest, mode_pairscale_fast, mode_pairscale)






# r_density_mode = function(x) {
#   d = stats::density(x, na.rm = TRUE)
#   return(d$x[which.max(d$y)])
# }
#
# N = 1000
# for(iter in 1:50) {
#   for(xnoise in seq(from = 2, to = 4, by = 0.1)) {
#     x = c(stats::rnorm(N), stats::rnorm(floor(N * 0.1), mean = xnoise), stats::rnorm(N*0.2, mean=0, sd = 0.1))
#     # plot(density(x, adjust = 1))
#     x_sans_outliers = x[x > stats::quantile(x, probs = 0.025, na.rm = TRUE) & x < stats::quantile(x, probs = 0.9975, na.rm = TRUE)]
#     mode_ref = r_density_mode(x_sans_outliers)
#     mode = density_mode(x, minval = 1, n_points = 512, adjust = 1, max_dist = 3, na_mode = "check")
#     # relatively high tolerance because our implementation is expected to return "robust mode" so not exactly the same computation
#     expect_equal(all.equal(mode, mode_ref, tolerance = 0.1), TRUE)
#   }
# }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{pairdiff_mean}
\alias{pairdiff_mean}
\title{Distance matrix between columns of a matrix using their mean differences}
\usage{
pairdiff_mean(x, cols = NULL, min_value_count = 3L, na_mode = "check")
}
\arguments{
\item{x}{a numeric matrix that needs to be normalized. This variable is changed by reference! i.e. after this function, the original variable is updated}

\item{cols}{optionally, provide an integer vector with column indices (in \code{x}) that should be used (these should be 1-based indices as per usual in R). Or set to \code{NULL} or \code{integer()} to use all columns}

\item{min_value_count}{the minimum number of values overlapping between a pair of columns (if there are fewer overlapping values, respective rescaling is not computed)}

\item{na_mode}{string value that indicated how should we should deal with NA values (default). "check" = test if NA values are present and if so, remove these. "present" = you already know NA values are present so we can skip the check for NA values for efficiency. "unchecked" = you guarantee that there are no NA values in input, thus we'll use the fastest code paths that skip over any NA checks downstream; ONLY select this if you are sure there are no NA/NaN/Inf/-Inf values !}
}
\value{
a N x N numeric matrix (where N is number of column in input \code{x}) representing the mean difference between each column
}
\description{
Pairwise differences between all columns in a matrix
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{pairscale_median}
\alias{pairscale_median}
\title{Normalize matrix columns using their median differences}
\usage{
pairscale_median(
  x,
  clusters = NULL,
  min_value_count = 3L,
  niter_irls = 50L,
  na_mode = "check"
)
}
\arguments{
\item{x}{a numeric matrix that needs to be normalized. This variable is changed by reference! i.e. after this function, the original variable is updated}

\item{clusters}{an integer vector, of same length as \code{ncol(x)}, that describes cluster identifiers. These values must describe a continuous set of integers between 1 and the total number of unique clusters. For example, a 6 column matrix with 3 columns from some experimental condition followed by 3 columns from another condition would be \code{c(1,1,1, 2,2,2)}. To disable mode-between rescaling, i.e. treat the entire matrix as one cluster, provide a vector with ones.}

\item{min_value_count}{the minimum number of values overlapping between a pair of columns (if there are fewer overlapping values, respective rescaling is not computed)}

\item{niter_irls}{number of iterations to increase robustness (set to zero to disable)}

\item{na_mode}{string value that indicated how should we should deal with NA values (default). "check" = test if NA values are present and if so, remove these. "present" = you already know NA values are present so we can skip the check for NA values for efficiency. "unchecked" = you guarantee that there are no NA values in input, thus we'll use the fastest code paths that skip over any NA checks downstream; ONLY select this if you are sure there are no NA/NaN/Inf/-Inf values !}
}
\value{
a numeric vector that represents the normalization factors that were applied to each column in x. Note that x is updated by reference.
}
\description{
Pairwise normalization of columns in a matrix, using the median to define pairwise distances between columns
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{vector_trimmedmean}
\alias{vector_trimmedmean}
\title{Compute trimmed mean value of a vector, with optional filtering for N datapoints}
\usage{
vector_trimmedmean(x, min_value_count = 1L, trim = 0.2, na_mode = "check")
}
\arguments{
\item{x}{numeric input vector, may contain non-finite values (removed if \code{na_mode} is left to default)}

\item{min_value_count}{the minimum number of values overlapping between a pair of columns (if there are fewer overlapping values, respective rescaling is not computed)}

\item{trim}{amount of trim to apply to both the lower- and upper-parts of a vector before computing the mean. 0 indicates no trim, 0.5 indicates 100\% trim (i.e. 50\% of data on both sides) so that value is out of bounds. Typically set to 0.1-0.3}

\item{na_mode}{string value that indicated how should we should deal with NA values (default). "check" = test if NA values are present and if so, remove these. "present" = you already know NA values are present so we can skip the check for NA values for efficiency. "unchecked" = you guarantee that there are no NA values in input, thus we'll use the fastest code paths that skip over any NA checks downstream; ONLY select this if you are sure there are no NA/NaN/Inf/-Inf values !}
}
\value{
a single numeric value representing the trimmed mean
}
\description{
Compute measure of central tendency using efficient C++ code
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pairscale.R
\docType{package}
\name{pairscale}
\alias{pairscale-package}
\alias{pairscale}
\title{pairscale package declaration}
\description{
Normalization of numerical matrices by minimizing the mean/median/mode difference between all column pairs.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ftwkoopmans/pairscale/}
}

}
\author{
\strong{Maintainer}: Frank Koopmans \email{ftwkoopmans@gmail.com} (\href{https://orcid.org/0000-0002-4973-5732}{ORCID})

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{solve_graph_laplacian}
\alias{solve_graph_laplacian}
\title{graph Laplacian approach to finding normalization factors}
\usage{
solve_graph_laplacian(M, niter_irls = 1L)
}
\arguments{
\item{M}{skew-symmetric input matrix, generated with e.g. \code{pairdiff_median()}}

\item{niter_irls}{refine the initial estimate using N additional iterative reweighted least squares loops for robust graph laplacian}
}
\value{
a numeric vector of length \code{ncol(M)} that contains scaling factors for \code{M}
}
\description{
find normalization factors for a given distance matrix computed with e.g. \code{pairdiff_median()}. For increased robustness, this function offers iterative reweighted improvement of the initial estimate.
}
\examples{
# toy example
x = cbind(
  c(1,2,3,4),
  c(2,3,4,9),
  c(1,2,4,5),
  c(1,0,1,0)
)
# compute pairwide median difference between all columns
M = pairscale::pairdiff_median(x)
# solve matrix M to find scaling factors, without and with reweighting
s1 = pairscale::solve_graph_laplacian(M, niter_irls = 0)
s2 = pairscale::solve_graph_laplacian(M, niter_irls = 10)
# rescaled matrices; only the robust variant correctly aligns columns 1 and 2
t(t(x) - s1)
t(t(x) - s2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{vector_mode}
\alias{vector_mode}
\title{Compute mode of a vector, with optional filtering for N datapoints}
\usage{
vector_mode(
  x,
  min_value_count = 3L,
  n_bins = 512L,
  adjust = 1,
  kernel_width_in_sd = 3,
  bandwidth_method = "nrd",
  mode_frac_maxdens = 1,
  na_mode = "check"
)
}
\arguments{
\item{x}{numeric input vector, may contain non-finite values (removed if \code{na_mode} is left to default)}

\item{min_value_count}{the minimum number of values overlapping between a pair of columns (if there are fewer overlapping values, respective rescaling is not computed)}

\item{n_bins}{grid size for density computation (the resolution / number of data points to use for binning column diffs)}

\item{adjust}{bandwidth adjust factor for density computation}

\item{kernel_width_in_sd}{maximum distance in standard deviations at which we'll include data points for the Gaussian kernal. Typically 3 or 4}

\item{bandwidth_method}{method in which this function computes bandwidth and optionally trims the data prior to binning. "nrd" is the robust, safe default. "nrd_fast" is faster and yields similar results for most distributions. Use "nrd_fastest" only when all pairwise distances are known to be near gaussian (i.e. no strong outliers and sd() is a reliable metric). "nrd_subset" is an experimental option that may be removed, it is fast but heavily favors symmetric distributions and is thus biased !

Valid options:
\itemize{
\item \code{"nrd_fastest"} ; use all data points, for bandwidth don't use IQR. Sensitive to outliers !
\item \code{"nrd_fast"} ; trim to quantiles 0.05 and 0.95, then use sd over these data points to compute bandwidth. Not as robust as the original nrd method, which also considers IQR, but much faster.
\item \code{"nrd"} ; adapted version of nrd; instead of IQR and sd we use IQR and sd over data trimmed at quantiles 0.05 and 0.95. Since we already sorted the data, computing sd on trimmed subset is 'free'
}}

\item{mode_frac_maxdens}{set to 1 to return the x-coordinate where the density is highest (mode). Setting this to a value < 1 will make this function compute not the mode, but the mean (x) value of the density where the density is some fraction higher than the maximum density. Typical value; 1. Optionally, set to 0.9 or 0.8 for possibly more robust center-finding, depending in your data distribution. Must not be smaller than 0.1 but recommended to never be lower than 0.7}

\item{na_mode}{string value that indicated how should we should deal with NA values (default). "check" = test if NA values are present and if so, remove these. "present" = you already know NA values are present so we can skip the check for NA values for efficiency. "unchecked" = you guarantee that there are no NA values in input, thus we'll use the fastest code paths that skip over any NA checks downstream; ONLY select this if you are sure there are no NA/NaN/Inf/-Inf values !}
}
\value{
a single numeric value representing the mode
}
\description{
Compute measure of central tendency using efficient C++ code
}
